// $Id$

//////////////////////////////////////////////////////////////////////////

#include "util.h"

#include <cv.h>
#include <highgui.h>
#include <imgptr.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <sstream>
#include <limits>

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 1
#endif

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

class prog_args
{
public:
  std::string in_name; // name of input file
  std::string out_name; // name of output file or directory

public:
  double color[3][2]; // color interval
  int dilate; // dilate the mask in given number of steps

public:
  prog_args(int argc, char * argv[]); // reads arguments
};

//////////////////////////////////////////////////////////////////////////

prog_args::prog_args(int argc, char * argv[])
{
  int const arg_num = argc - 1;
  int const arg_num_min = 8;
  int const arg_num_max = arg_num_min + 1;

  if(arg_num < arg_num_min || arg_num > arg_num_max)
  {
    std::ostringstream error;
    error
      << "arguments:"
      << " <in_name> <out_name>"
      << " <r_low> <r_up> <g_low> <g_up> <b_low> <b_up>"
      << " [dilate]"
    ;
    throw std::runtime_error(error.str());
  }

  // MANDATORY ARGUMENTS

  in_name = argv[1]; // name of input file
  out_name = argv[2]; // name of output file or directory

  color[0][0] = atof(argv[3]); // color interval
  color[0][1] = atof(argv[4]);

  color[1][0] = atof(argv[5]);
  color[1][1] = atof(argv[6]);

  color[2][0] = atof(argv[7]);
  color[2][1] = atof(argv[8]);

  for(int a = 0; a < 3; ++a)
  {
    for(int b = 0; b < 2; ++b)
    {
      color[a][b] *= 255;
    }
  }

  // OPTIONAL ARGUMENTS

  dilate =
    (arg_num >= arg_num_min + 1)
    ? atoi(argv[arg_num_min + 1])
    : 0; // do not dilate
}

//////////////////////////////////////////////////////////////////////////

class prog
{
private:
  prog_args const _args; // program arguments
  cv::img_ptr _in_img, _out_img; // input /output image

public:
  prog(int argc, char * argv[])
  :
    _args (argc, argv) // reads arguments
  {
    _init(); // initializes the program
    _run(); // executes the program
  }

private:
  prog(prog const &); // NOT IMPLEMENTED
  prog& operator=(prog const &);

private:
  void _init(); // initializes the program
  void _run(); // executes the program
};

//////////////////////////////////////////////////////////////////////////

void prog::_init()
{
  // OPEN INPUT

  _in_img = cvLoadImage(_args.in_name.c_str());

  if(_in_img == 0 || _in_img->imageData == 0)
  {
    std::ostringstream error;
    error << "couldn't load file: " << _args.in_name;
    throw std::runtime_error(error.str());
  }

  if(_in_img->nChannels != 3) // BGR
  {
    std::ostringstream error;
    error
      << "unsupported number of input channels: "
      << _in_img->nChannels;
    throw std::runtime_error(error.str());
  }

  if(_in_img->depth != (int) IPL_DEPTH_8U) // [0,255]
  {
    std::ostringstream error;
    error
      << "unsupported type of input data: "
      << ipl_depth_name(_in_img->depth);
    throw std::runtime_error(error.str());
  }

  if(_in_img->width < MIN_FSIZE || _in_img->height < MIN_FSIZE)
  {
    std::ostringstream error;
    error
      << "too small input image: "
      << _in_img->width << "x" << _in_img->height
      << " (must be >= " << MIN_FSIZE << ")";
    throw std::runtime_error(error.str());
  }

  CvSize const in_sz = cvSize(_in_img->width, _in_img->height);

#if TEST_LEVEL >= 1
  std::cerr
    << "Input: " << _args.in_name << " "
    << in_sz.width << "x" << in_sz.height << " "
    << std::endl;
#endif

  // OPEN OUTPUT

  CvSize const out_sz = in_sz; // output size == input size

#if TEST_LEVEL >= 1
  std::cerr
    << "Output: " << _args.out_name << " "
    << out_sz.width << "x" << out_sz.height << " "
    << std::endl;
#endif

  _out_img.reset(out_sz, CV_8UC1); // output image
}

//////////////////////////////////////////////////////////////////////////

void prog::_run()
{
  CvScalar const ref_lower =
    cvScalar(_args.color[2][0], _args.color[1][0], _args.color[0][0]);

  CvScalar const ref_upper =
    cvScalar(_args.color[2][1], _args.color[1][1], _args.color[0][1]);

  cvInRangeS(_in_img, ref_lower, ref_upper, _out_img);

  if(_args.dilate > 0)
  {
    cvDilate(_out_img, _out_img, 0, _args.dilate);
    cvSmooth(_out_img, _out_img, CV_MEDIAN);
  }

  cvNot(_out_img, _out_img);

  cvSaveImage(_args.out_name.c_str(), _out_img);
}

//////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv[])
{
  std::string const prog_name = argv[0];
  try
  {
    prog(argc, argv);
    return EXIT_SUCCESS;
  }
  catch(std::exception const & e)
  {
    std::cerr
      << prog_name << ": " << e.what()
      << std::endl;
  }
  return EXIT_FAILURE;
}

//////////////////////////////////////////////////////////////////////////

// End of File
