// $Id$

//////////////////////////////////////////////////////////////////////////

#include "inpaint.h"

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 0
#endif

#ifndef MIN_TEST_LEVEL
#define MIN_TEST_LEVEL 2
#endif

//////////////////////////////////////////////////////////////////////////

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

#include <cv.h>
#include <imgptr.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <cmath>
#include <vector>
#include <new>

//////////////////////////////////////////////////////////////////////////

#if TEST_LEVEL >= MIN_TEST_LEVEL
#include <highgui.h>
#include <iostream>
#include <sstream>
#endif

//////////////////////////////////////////////////////////////////////////
namespace _mm_inp_ {
//////////////////////////////////////////////////////////////////////////

inline
bool operator==(CvSize const & a, CvSize const & b)
{
  return (a.width == b.width) && (a.height == b.height);
}

inline
bool operator!=(CvSize const & a, CvSize const & b)
{
  return !(a == b);
}

//////////////////////////////////////////////////////////////////////////

typedef std::vector<CvSize> size_array;
typedef std::vector<float> real_array;

typedef cv::img_ptr img_ptr;
typedef std::vector<img_ptr> img_array;
typedef std::vector<img_array> multi_img_array;

//////////////////////////////////////////////////////////////////////////
} // namespace _mm_inp_
//////////////////////////////////////////////////////////////////////////

using namespace _mm_inp_;

//////////////////////////////////////////////////////////////////////////

class _mm_inp // Multiscale Multichannel Inpainting
{
protected:
  int _lev_num; // number of scale levels
  double _lambda; // scaling factor
  CvSize _img_size; // can handle images of this size
  int _channels; // number of input channels

protected:
  size_array _size_pyr; // size on different levels
  real_array _channel_weights; // weights of input channels
  real_array _scaling_exponents; // sacaling of weights

protected:
  img_ptr _io_img; // input/output temporary data (multi-channeled)

  multi_img_array _img_pyr; // image pyramid (input/output, split)
  multi_img_array _ups_img_pyr; // upscaled image (switching levels)
  multi_img_array _dat_pyr[2]; // temporary data (diffusion)

  img_array _norm_pyr; // norm of weighted image components
  img_array _norm_gr_pyr[2]; // gradient of norm (diffusivity)
  img_array _diff_pyr[3]; // aniostropic diffusivity tensor

  img_array _mask_pyr; // mask pyramid (input, single channel)

public:
  _mm_inp()
  :
    _lev_num (0), _lambda (0),
    _img_size (cvSize(0, 0)),
    _channels (0)
  {
  }

protected:
  void _drop_data(); // deallocates memory

  void _fill_img_pyr(CvArr const * img, CvArr const * mask);
  void _fill_mask_pyr(CvArr const * mask);

  void _init_data(
    CvArr const * img, CvArr const * mask, int lev_num, double lambda,
    double const * channel_weights, double const * scaling_exponents
  );

protected:
  void _correct_boundary(
    cv::img_ptr & img, // REPLICATES VALUES on bottom-right boundary
    bool is_x_corr, bool is_y_corr
  );

  void _switch_to_level(int lev); // FROM LOWER TO HIGHER
  void _run_on_level(int lev, int it_num, double lambda, double alpha);
  void _init_output(CvArr * out);

public:
  void inpaint(
    CvArr const * input, CvArr const * mask, CvArr * output,
    int it_num, int lev_num, double lambda, double alpha, // parameters
    double const * channel_weights, double const * scaling_exponents
  );
};

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_drop_data()
{
  _lev_num = 0;
  _lambda = 0;
  _img_size = cvSize(0, 0);
  _channels = 0;

  _size_pyr.clear();
  _channel_weights.clear();
  _scaling_exponents.clear();

  _io_img.reset();

  _img_pyr.clear();
  _ups_img_pyr.clear();
  _dat_pyr[0].clear();
  _dat_pyr[1].clear();

  _norm_pyr.clear();
  _norm_gr_pyr[0].clear();
  _norm_gr_pyr[1].clear();
  _diff_pyr[0].clear();
  _diff_pyr[1].clear();
  _diff_pyr[2].clear();

  _mask_pyr.clear();
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_fill_img_pyr(CvArr const * img, CvArr const * mask)
{
  // init the highest level

  {
    int const lev = 0; // HIGHEST level
    cvConvert(img, _io_img); // makes a copy, CONVERTS to float

  #if 0 // HACK
    cvSet(_io_img, cvScalarAll(1), mask);
  #endif

    #pragma omp parallel for
    for(int y = 0; y < _img_size.height; ++y) // SPLITS into CHANNELS
    {
      float const * const img_row = static_cast<float *>(_io_img.row(y));

      float * I_row [_channels]; // image channels
      for(int k = 0; k < _channels; ++k)
      {
        I_row[k] = static_cast<float *>(_img_pyr[lev][k].row(y));
      }

      for(int x = 0; x < _img_size.width; ++x)
      {
        for(int k = 0; k < _channels; ++k)
        {
          I_row[k][x] = img_row[_channels * x + k];
        }
      }
    }
  }

  // downscale

  if(_lambda == 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      for(int k = 0; k < _channels; ++k)
      {
        cvPyrDown(_img_pyr[lev-1][k], _img_pyr[lev][k]); // HALF size
      }
    }
  }
  else // if(_lambda != 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      for(int k = 0; k < _channels; ++k)
      {
        cvResize(
          _img_pyr[lev-1][k], _img_pyr[lev][k],
          //CV_INTER_NN       // Nearest-neigbor interpolation
          CV_INTER_LINEAR   // Bilinear interpolation
          //CV_INTER_AREA     // Resampling using pixel area relation
          //CV_INTER_CUBIC    // Bicubic interpolation
        );
      }
    }
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_fill_mask_pyr(CvArr const * mask)
{
  // init the highest level

  {
    int const lev = 0; // HIGHEST level
    cvCopy(mask, _mask_pyr[lev]);
  }

  // downscale

  if(_lambda == 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      cvPyrDown(_mask_pyr[lev-1], _mask_pyr[lev]); // HALF size
    }
  }
  else // if(_lambda != 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      cvResize(
        _mask_pyr[lev-1], _mask_pyr[lev],
        //CV_INTER_NN       // Nearest-neigbor interpolation
        CV_INTER_LINEAR   // Bilinear interpolation
        //CV_INTER_AREA     // Resampling using pixel area relation
        //CV_INTER_CUBIC    // Bicubic interpolation
      );
    }
  }

  // threshold smoothed values

  for(int lev = 0; lev < _lev_num; ++lev)
  {
    cvThreshold(
      _mask_pyr[lev], _mask_pyr[lev],
      128, 255, CV_THRESH_BINARY
    );
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_init_data(
  CvArr const * img, CvArr const * mask, int lev_num, double lambda,
  double const * channel_weights, double const * scaling_exponents
)
{
  assert(img != 0 && mask != 0);
  assert(lev_num > 0); // AT LEAST ONE LEVEL needed
  assert(lambda > 0 && lambda < 1);

  CvSize const img_size = cvGetSize(img);

  assert(img_size.width >= MIN_FSIZE); // avoid too small images
  assert(img_size.height >= MIN_FSIZE);

  assert(cvGetSize(mask) == img_size);

  int const channels = CV_MAT_CN(cvGetElemType(img));
  assert(channels > 0); // AT LEAST ONE CHANNEL needed

  assert(cvGetElemType(mask) == CV_8UC1);
  //assert(CV_MAT_CN(cvGetElemType(mask)) == 1);

  if(
    _lev_num != lev_num || _lambda != lambda ||
    _img_size != img_size || _channels != channels
  )
  {
    _drop_data(); // DROPS OLD DATA, if there was any

    // create pyramids

    _img_pyr.resize(lev_num); // input images
    _ups_img_pyr.resize(lev_num);
    _dat_pyr[0].resize(lev_num);
    _dat_pyr[1].resize(lev_num);

    _norm_pyr.resize(lev_num);
    _norm_gr_pyr[0].resize(lev_num);
    _norm_gr_pyr[1].resize(lev_num);
    _diff_pyr[0].resize(lev_num);
    _diff_pyr[1].resize(lev_num);
    _diff_pyr[2].resize(lev_num);

    _mask_pyr.resize(lev_num);

    for(int lev = 0; lev < lev_num; ++lev)
    {
      _img_pyr[lev].resize(channels);
      _ups_img_pyr[lev].resize(channels);
      _dat_pyr[0][lev].resize(channels);
      _dat_pyr[1][lev].resize(channels);
    }

    // calculate image size on pyramid levels

    {
      _size_pyr.resize(lev_num);

      CvSize sz = img_size; // HIGHEST LEVEL
      for(int lev = 0; lev < lev_num; ++lev)
      {
        assert(sz.width >= MIN_FSIZE && sz.height >= MIN_FSIZE);
        _size_pyr[lev] = sz; // image size on current level

        if(lambda == 0.5)
        {
          sz.width /= 2, sz.height /= 2; // HALF size on next level
        }
        else // if(lambda != 0.5)
        {
          sz.width = static_cast<int>(floor(sz.width * lambda));
          sz.height = static_cast<int>(floor(sz.height * lambda));
        }
      }
    }

    // create images

    _io_img.reset(img_size, IPL_DEPTH_32F, channels);

    for(int lev = 0; lev < lev_num; ++lev)
    {
      CvSize const sz = _size_pyr[lev]; // current size

      for(int k = 0; k < channels; ++k)
      {
        _img_pyr[lev][k].reset(sz, IPL_DEPTH_32F, 1);
        _ups_img_pyr[lev][k].reset(sz, IPL_DEPTH_32F, 1);
        _dat_pyr[0][lev][k].reset(sz, IPL_DEPTH_32F, 1);
        _dat_pyr[1][lev][k].reset(sz, IPL_DEPTH_32F, 1);
      }

      _norm_pyr[lev].reset(sz, IPL_DEPTH_32F, 1);
      _norm_gr_pyr[0][lev].reset(sz, IPL_DEPTH_32F, 1);
      _norm_gr_pyr[1][lev].reset(sz, IPL_DEPTH_32F, 1);
      _diff_pyr[0][lev].reset(sz, IPL_DEPTH_32F, 1);
      _diff_pyr[1][lev].reset(sz, IPL_DEPTH_32F, 1);
      _diff_pyr[2][lev].reset(sz, IPL_DEPTH_32F, 1);

      _mask_pyr[lev].reset(sz, IPL_DEPTH_8U, 1);
    }

    // store parameters of pyramids

    _lev_num = lev_num;
    _lambda = lambda;
    _img_size = img_size;
    _channels = channels;
  }

  // init channel weights and scaling exponents

  if(channel_weights == 0)
  {
    _channel_weights.resize(_channels, 1); // all weights == 1
  }
  else // if(channel_weights != 0)
  {
    _channel_weights.resize(_channels);

    for(int k = 0; k < _channels; ++k)
    {
      _channel_weights[k] = channel_weights[k];
    }
  }

  if(scaling_exponents == 0)
  {
    _scaling_exponents.resize(_channels, 0); // all exponents == 1
  }
  else // if(scaling_exponents != 0)
  {
    _scaling_exponents.resize(_channels);

    for(int k = 0; k < _channels; ++k)
    {
      _scaling_exponents[k] = scaling_exponents[k];
    }
  }

  // init image and mask pyramid

  _fill_img_pyr(img, mask);
  _fill_mask_pyr(mask);
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_correct_boundary(
  cv::img_ptr & img, // REPLICATES VALUES on bottom-right boundary
  bool is_x_corr, bool is_y_corr
)
{
  CvSize const sz = cvGetSize(img);

  if(is_x_corr) // RIGHT column
  {
    for(int y = 0; y < sz.height; ++y)
    {
      float * const row =
        reinterpret_cast<float *>(img.row(y));

      row[sz.width-1] = row[sz.width-2];
    }
  }

  if(is_y_corr) // BOTTOM row
  {
    float * const last_row =
      reinterpret_cast<float *>(img.row(sz.height-1));
    float const * const curr_row =
      reinterpret_cast<float const *>(img.row(sz.height-2));

    memcpy(last_row, curr_row, sz.width * sizeof(float));
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_switch_to_level(int lev)
{
  assert(lev < _lev_num - 1); // FROM LOWER TO HIGHER

  if(_lambda == 0.5)
  {
    CvSize const sz = _size_pyr[lev];
    CvSize const sz_lower = _size_pyr[lev+1];

    CvSize const sz_upper =
      cvSize(2 * sz_lower.width, 2 * sz_lower.height);

    bool const is_x_corr = (sz.width != sz_upper.width);
    bool const is_y_corr = (sz.height != sz_upper.height);

    bool const is_bound_corr_needed = is_x_corr || is_y_corr;

    CvRect const rect_lower_up =
      cvRect(0, 0, sz_upper.width, sz_upper.height);

    for(int k = 0; k < _channels; ++k)
    {
      if(is_bound_corr_needed) // boundary correction?
      {
        cvSetImageROI(_ups_img_pyr[lev][k], rect_lower_up);
      }

      cvPyrUp(_img_pyr[lev+1][k], _ups_img_pyr[lev][k]); // size DOUBLED

      if(is_bound_corr_needed) // boundary correction?
      {
        cvResetImageROI(_ups_img_pyr[lev][k]); // resets region of interest
        _correct_boundary(_ups_img_pyr[lev][k], is_x_corr, is_y_corr);
      }
    }
  }
  else // if(_lambda != 0.5)
  {
    for(int k = 0; k < _channels; ++k)
    {
      cvResize(
        _img_pyr[lev+1][k], _ups_img_pyr[lev][k],
        //CV_INTER_NN       // Nearest-neigbor interpolation
        CV_INTER_LINEAR   // Bilinear interpolation
        //CV_INTER_AREA     // Resampling using pixel area relation
        //CV_INTER_CUBIC    // Bicubic interpolation
      );
    }
  }

  for(int k = 0; k < _channels; ++k)
  {
    cvCopy(_ups_img_pyr[lev][k], _img_pyr[lev][k], _mask_pyr[lev]);
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_run_on_level(
  int lev, int it_num, double lambda, double alpha
)
{
  assert(lev < _lev_num);

#if 0 // HACK
#if 0
  int const der_kern = CV_SCHARR;
  double const der_scale = 1.0 / 32.0;
#else
  int const der_kern = 3; // Sobel
  double const der_scale = 1.0 / 8.0;
#endif
#else
  int const der_kern = 1; // kernel: -1 0 1
  double const der_scale = 1.0 / 2.0;
#endif

  CvSize const sz = _size_pyr[lev]; // size on current level

  for(int step = 0; step < it_num; ++step)
  {
    // calculate norm image

    #pragma omp parallel for
    for(int y = 0; y < sz.height; ++y)
    {
      float const * I_row [_channels]; // image channels
      for(int k = 0; k < _channels; ++k)
      {
        I_row[k] = static_cast<float *>(_img_pyr[lev][k].row(y));
      }

      float * const norm_row =
        static_cast<float *>(_norm_pyr[lev].row(y));

      for(int x = 0; x < sz.width; ++x)
      {
      #if 0 // HACK
        float weighted_sum = 0;

        for(int k = 0; k < _channels; ++k)
        {
          float const img_val = I_row[k][x];

          float const exponent = lev * _scaling_exponents[k];
          float const scaling = pow(_lambda, exponent);
          float const weight = scaling * _channel_weights[k];

          weighted_sum += (weight * img_val* img_val);
        }

        norm_row[x] = sqrt(weighted_sum); // Euclidean norm
      #else
        float weighted_sum = 0;

        for(int k = 0; k < _channels; ++k)
        {
          float const img_val = I_row[k][x];

          float const exponent = lev * _scaling_exponents[k];
          float const scaling = pow(_lambda, exponent);
          float const weight = scaling * _channel_weights[k];

          weighted_sum += fabs(weight * img_val);
        }

        norm_row[x] = weighted_sum; // L1 norm
      #endif
      }
    }

    // calculate norm gradients

    cvSobel(_norm_pyr[lev], _norm_gr_pyr[0][lev], 1, 0, der_kern);
    cvSobel(_norm_pyr[lev], _norm_gr_pyr[1][lev], 0, 1, der_kern);

    cvScale(_norm_gr_pyr[0][lev], _norm_gr_pyr[0][lev], der_scale);
    cvScale(_norm_gr_pyr[1][lev], _norm_gr_pyr[1][lev], der_scale);

    // calculate diffusivity

    #pragma omp parallel for
    for(int y = 0; y < sz.height; ++y)
    {
      float const * const norm_gr_row [2] = // norm gradients
      {
        static_cast<float *>(_norm_gr_pyr[0][lev].row(y)),
        static_cast<float *>(_norm_gr_pyr[1][lev].row(y)),
      };

      float * const D_row [3] = // diffusivity
      {
        static_cast<float *>(_diff_pyr[0][lev].row(y)),
        static_cast<float *>(_diff_pyr[1][lev].row(y)),
        static_cast<float *>(_diff_pyr[2][lev].row(y)),
      };

      for(int x = 0; x < sz.width; ++x)
      {
        float const gr[2] = { norm_gr_row[0][x], norm_gr_row[1][x] };

        float const gr_sc_thr = 1e-8; // THRESHOLD
        float const gr_nr_sc = gr[0] * gr[0] + gr[1] * gr[1];
        float const gr_sc = (gr_nr_sc < gr_sc_thr) ? 0 : 1 / gr_nr_sc;

        float const eta1 = alpha;
        float const eta2 = 0e-2 * alpha; // HACK

        float const eta_diff = eta1 - eta2;

        D_row[0][x] = eta1 - eta_diff * gr_sc * gr[0] * gr[0];
        D_row[1][x] = eta1 - eta_diff * gr_sc * gr[1] * gr[1];
        D_row[2][x] =      - eta_diff * gr_sc * gr[0] * gr[1];
      }
    }

    // smooth diffusivity

    cvSmooth(_diff_pyr[0][lev], _diff_pyr[0][lev], CV_GAUSSIAN, 3);
    cvSmooth(_diff_pyr[1][lev], _diff_pyr[1][lev], CV_GAUSSIAN, 3);
    cvSmooth(_diff_pyr[2][lev], _diff_pyr[2][lev], CV_GAUSSIAN, 3);

    // calculate image gradients

    for(int k = 0; k < _channels; ++k)
    {
      cvSobel(_img_pyr[lev][k], _dat_pyr[0][lev][k], 1, 0, der_kern);
      cvSobel(_img_pyr[lev][k], _dat_pyr[1][lev][k], 0, 1, der_kern);

      cvScale(_dat_pyr[0][lev][k], _dat_pyr[0][lev][k], der_scale);
      cvScale(_dat_pyr[1][lev][k], _dat_pyr[1][lev][k], der_scale);
    }

    // update image gradients based on diffusivity

    #pragma omp parallel for
    for(int y = 0; y < sz.height; ++y)
    {
      float * gr_row [2][_channels]; // image gradients
      for(int k = 0; k < _channels; ++k)
      {
        gr_row[0][k] = static_cast<float *>(_dat_pyr[0][lev][k].row(y));
        gr_row[1][k] = static_cast<float *>(_dat_pyr[1][lev][k].row(y));
      }

      float const * const D_row [3] = // diffusivity
      {
        static_cast<float *>(_diff_pyr[0][lev].row(y)),
        static_cast<float *>(_diff_pyr[1][lev].row(y)),
        static_cast<float *>(_diff_pyr[2][lev].row(y)),
      };

      for(int x = 0; x < sz.width; ++x)
      {
        float const D[3] = { D_row[0][x], D_row[1][x], D_row[2][x] };

        for(int k = 0; k < _channels; ++k)
        {
          float const gr[2] = { gr_row[0][k][x], gr_row[1][k][x] };

          gr_row[0][k][x] = D[0] * gr[0] + D[2] * gr[1];
          gr_row[1][k][x] = D[2] * gr[0] + D[1] * gr[1];
        }
      }
    }

    // calculate gradient of gradients

    for(int k = 0; k < _channels; ++k)
    {
      cvSobel(_dat_pyr[0][lev][k], _dat_pyr[0][lev][k], 1, 0, der_kern);
      cvSobel(_dat_pyr[1][lev][k], _dat_pyr[1][lev][k], 0, 1, der_kern);

      cvScale(_dat_pyr[0][lev][k], _dat_pyr[0][lev][k], der_scale);
      cvScale(_dat_pyr[1][lev][k], _dat_pyr[1][lev][k], der_scale);
    }

    // update image within mask

    for(int k = 0; k < _channels; ++k)
    {
      cvAdd(
        _img_pyr[lev][k], _dat_pyr[0][lev][k], _img_pyr[lev][k],
        _mask_pyr[lev]
      );
      cvAdd(
        _img_pyr[lev][k], _dat_pyr[1][lev][k], _img_pyr[lev][k],
        _mask_pyr[lev]
      );
    }
  }

#if TEST_LEVEL >= MIN_TEST_LEVEL + 1
  //cvNamedWindow("run");
  //cvShowImage("run", _img_pyr[lev][0]);
  cvShowImage("mask", _mask_pyr[lev]);
  cvWaitKey(10);
#endif
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_init_output(CvArr * out)
{
  assert(out != 0 && cvGetSize(out) == _img_size);
  int const lev = 0; // HIGHEST level

  #pragma omp parallel for
  for(int y = 0; y < _img_size.height; ++y) // JOINS CHANNELS
  {
    float const * I_row [_channels]; // image channels
    for(int k = 0; k < _channels; ++k)
    {
      I_row[k] = static_cast<float const *>(_img_pyr[lev][k].row(y));
    }

    float * const out_row = static_cast<float *>(_io_img.row(y));

    for(int x = 0; x < _img_size.width; ++x)
    {
      float * const out_pix = out_row + _channels * x;

      for(int k = 0; k < _channels; ++k)
      {
        out_pix[k] = I_row[k][x];
      }
    }
  }

  cvConvert(_io_img, out); // CONVERTS from float to output type
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::inpaint(
  CvArr const * input, CvArr const * mask, CvArr * output,
  int it_num, int lev_num, double lambda, double alpha, // parameters
  double const * channel_weights, double const * scaling_exponents
)
{
  _init_data(
    input, mask, lev_num, lambda,
    channel_weights, scaling_exponents
  );

  int lev = _lev_num - 1; // staring on the LOWEST level

#if TEST_LEVEL >= MIN_TEST_LEVEL
  {
    CvSize const & size_k = _size_pyr[lev];
    std::cerr
      << "Level: "
      << size_k.width << " x " << size_k.height
      << std::endl;
  }
#endif

  while(lev > 0) // from lower to higher
  {
    _run_on_level(lev, it_num, lambda, alpha); // CURRENT level

    --lev; // switching TO HIGHER LEVEL

  #if TEST_LEVEL >= MIN_TEST_LEVEL
    {
      CvSize const & size_k = _size_pyr[lev];
      std::cerr
        << "Level: "
        << size_k.width << " x " << size_k.height
        << std::endl;
    }
  #endif

    _switch_to_level(lev); // moves results to HIGHER level
  }

  _run_on_level(0, it_num, lambda, alpha); // HIGHEST level
  _init_output(output);
}

//////////////////////////////////////////////////////////////////////////
// EXPORT
//////////////////////////////////////////////////////////////////////////

mm_inpaint_func::mm_inpaint_func()
:
  _intern (0)
{
  // NO MEMORY ALLOCATED
}

//////////////////////////////////////////////////////////////////////////

mm_inpaint_func::~mm_inpaint_func()
{
  reset(); // DEALLOCATES MEMORY, if there was any allocated
}

//////////////////////////////////////////////////////////////////////////

mm_inpaint_func::mm_inpaint_func(
  mm_inpaint_func const & func
)
:
  _intern (0)
{
  // there is NO REASON TO COPY anything
  // NO MEMORY ALLOCATED
}

//////////////////////////////////////////////////////////////////////////

mm_inpaint_func &
  mm_inpaint_func::operator=(mm_inpaint_func const & func)
{
  // there is NO REASON TO COPY anything
  return *this;
}

//////////////////////////////////////////////////////////////////////////

void mm_inpaint_func::reset()
{
  if(_intern == 0) return; // IGNORED
  delete static_cast<_mm_inp *>(_intern);
  _intern = 0;
}

//////////////////////////////////////////////////////////////////////////

void mm_inpaint_func::operator()(
  CvArr const * input, CvArr const * mask, CvArr * output,
  int it_num, int lev_num, double lambda, double alpha, // parameters
  double const * channel_weights, double const * scaling_exponents
)
{
  if(_intern == 0)
  {
    _intern = new _mm_inp ();
    assert(_intern != 0);
  }

  _mm_inp * intern = static_cast<_mm_inp *>(_intern);

  intern->inpaint(
    input, mask, output, it_num, lev_num, lambda,
    alpha, channel_weights, scaling_exponents
  );
}

//////////////////////////////////////////////////////////////////////////

// End of File
