// $Id$

//////////////////////////////////////////////////////////////////////////

#include "util.h"

//////////////////////////////////////////////////////////////////////////

#include <cassert>
#include <cmath>
#include <string>
#include <iterator>

//////////////////////////////////////////////////////////////////////////

bool is_ipl_depth_valid(int const depth)
{
  switch(depth)
  {
  case IPL_DEPTH_8U:
  case IPL_DEPTH_8S:
  case IPL_DEPTH_16U:
  case IPL_DEPTH_16S:
  //case IPL_DEPTH_32U:
  case IPL_DEPTH_32S:
  case IPL_DEPTH_32F:
  case IPL_DEPTH_64F:
    return true;
  default:
    ;
  }
  return false;
}

//////////////////////////////////////////////////////////////////////////

char const * ipl_depth_name(int const depth)
{
  switch(depth)
  {
  case IPL_DEPTH_8U:  return "8 bits unsigned fixed";
  case IPL_DEPTH_8S:  return "8 bits signed fixed";
  case IPL_DEPTH_16U: return "16 bits unsigned fixed";
  case IPL_DEPTH_16S: return "16 bits signed fixed";
  //case IPL_DEPTH_32U: return "32 bits unsigned fixed"; // not supported
  case IPL_DEPTH_32S: return "32 bits signed fixed";
  case IPL_DEPTH_32F: return "32 bits floating point";
  case IPL_DEPTH_64F: return "64 bits floating point";
  default:
    ;
  }
  return "undefined";
}

//////////////////////////////////////////////////////////////////////////

void cart_to_spheric(const CvArr * src, CvArr * dst)
{
  // changing Cartesian coordinates to Spherical

  assert(src != 0 && dst != 0);

  assert(cvGetElemType(src) == CV_32FC3); // Cartesian (in)
  assert(cvGetElemType(dst) == CV_32FC3); // Spherical (out)

  CvSize const sz = cvGetSize(src);
  assert(cvGetSize(dst) == sz);

  CvMat src_mat, dst_mat; // temporary headers
  int src_coi, dst_coi; // channel of interst not supported

  CvMat const * const src_ = cvGetMat(src, &src_mat, &src_coi);
  CvMat const * const dst_ = cvGetMat(dst, &dst_mat, &dst_coi);

  assert(src_ != 0 && dst_ != 0);
  assert(!src_coi && !dst_coi);

  #pragma omp parallel for
  for(int y = 0; y < sz.height; ++y)
  {
    float const * const src_row =
      reinterpret_cast<float const *>(src_->data.ptr + y * src_->step);

    float * const dst_row =
      reinterpret_cast<float *>(dst_->data.ptr + y * dst_->step);

    for(int x = 0; x < sz.width; ++x)
    {
      float const * const src_pix = src_row + 3 * x;
      float * const dst_pix = dst_row + 3 * x;

      float const X = src_pix[0]; // Cartesian coordinates
      float const Y = src_pix[1];
      float const Z = src_pix[2];

      float const r_XY_sq = X * X + Y * Y;

      float const r = sqrt(r_XY_sq + Z * Z);
      float const theta = atan2(sqrt(r_XY_sq), Z);
      float const phi = atan2(Y, X);

      dst_pix[0] = r; // distance (radius) from origin
      dst_pix[1] = theta; // vertical polar angle from the Z-axis
      dst_pix[2] = phi; // azimuthal angle in XY-plane from the X-axis
    }
  }
}

//////////////////////////////////////////////////////////////////////////

void spheric_to_cart(const CvArr * src, CvArr * dst)
{
  // changing Cartesian coordinates to Spherical

  assert(src != 0 && dst != 0);

  assert(cvGetElemType(src) == CV_32FC3); // Cartesian (in)
  assert(cvGetElemType(dst) == CV_32FC3); // Spherical (out)

  CvSize const sz = cvGetSize(src);
  assert(cvGetSize(dst) == sz);

  CvMat src_mat, dst_mat; // temporary headers
  int src_coi, dst_coi; // channel of interst not supported

  CvMat const * const src_ = cvGetMat(src, &src_mat, &src_coi);
  CvMat const * const dst_ = cvGetMat(dst, &dst_mat, &dst_coi);

  assert(src_ != 0 && dst_ != 0);
  assert(!src_coi && !dst_coi);

  #pragma omp parallel for
  for(int y = 0; y < sz.height; ++y)
  {
    float const * const src_row =
      reinterpret_cast<float const *>(src_->data.ptr + y * src_->step);

    float * const dst_row =
      reinterpret_cast<float *>(dst_->data.ptr + y * dst_->step);

    for(int x = 0; x < sz.width; ++x)
    {
      float const * const src_pix = src_row + 3 * x;
      float * const dst_pix = dst_row + 3 * x;

      float const r = src_pix[0]; // spherical coordinates
      float const theta = src_pix[1];
      float const phi = src_pix[2];

      float const r_XY = r * sin(theta);

      float const X = r_XY * cos(phi);
      float const Y = r_XY * sin(phi);
      float const Z = r * cos(theta);

      dst_pix[0] = X; // Cartesian coordinates
      dst_pix[1] = Y;
      dst_pix[2] = Z;
    }
  }
}

//////////////////////////////////////////////////////////////////////////

// End of File
