// $Id$

//////////////////////////////////////////////////////////////////////////

#include "inpaint.h"

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 0
#endif

#ifndef MIN_TEST_LEVEL
#define MIN_TEST_LEVEL 2
#endif

//////////////////////////////////////////////////////////////////////////

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

#include <cv.h>
#include <imgptr.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <cmath>
#include <vector>
#include <new>

//////////////////////////////////////////////////////////////////////////

#if TEST_LEVEL >= MIN_TEST_LEVEL
#include <highgui.h>
#include <iostream>
#include <sstream>
#endif

//////////////////////////////////////////////////////////////////////////
namespace _mm_inp_ {
//////////////////////////////////////////////////////////////////////////

inline
bool operator==(CvSize const & a, CvSize const & b)
{
  return (a.width == b.width) && (a.height == b.height);
}

inline
bool operator!=(CvSize const & a, CvSize const & b)
{
  return !(a == b);
}

//////////////////////////////////////////////////////////////////////////

typedef std::vector<CvSize> size_array;

typedef cv::img_ptr img_ptr;
typedef std::vector<img_ptr> img_array;
typedef std::vector<img_array> multi_img_array;

//////////////////////////////////////////////////////////////////////////
} // namespace _mm_inp_
//////////////////////////////////////////////////////////////////////////

using namespace _mm_inp_;

//////////////////////////////////////////////////////////////////////////

class _mm_inp // Multiscale Multichannel Inpainting
{
protected:
  int _lev_num; // number of scale levels
  double _lambda; // scaling factor
  CvSize _img_size; // can handle images of this size
  int _channels; // number of input channels

protected:
  size_array _size_pyr; // size on different levels

protected:
  img_ptr _tmp_img; // temporary data (multi-channeled)
  multi_img_array _img_pyr; // image pyramid (input/output, split)
  multi_img_array _tmp_img_pyr; // temporary data (switching levels)
  img_array _mask_pyr; // mask pyramid (input, single channel)

public:
  _mm_inp()
  :
    _lev_num (0), _lambda (0),
    _img_size (cvSize(0, 0)),
    _channels (0)
  {
  }

protected:
  void _drop_data(); // deallocates memory

  void _fill_img_pyr(CvArr const * img); // Gaussian pyramid
  void _fill_mask_pyr(CvArr const * mask);

  void _init_data(
    CvArr const * img, CvArr const * mask,
    int lev_num, double lambda
  );

protected:
  void _correct_boundary(
    cv::img_ptr & img, // REPLICATES VALUES on bottom-right boundary
    bool is_x_corr, bool is_y_corr
  );

  void _switch_to_level(int lev); // FROM LOWER TO HIGHER
  void _run_on_level(int lev, int it_num, double lambda, double alpha);
  void _init_output(CvArr * out);

public:
  void inpaint(
    CvArr const * input, CvArr const * mask, CvArr * output,
    int it_num, int lev_num, double lambda, double alpha
  );
};

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_drop_data()
{
  _lev_num = 0;
  _lambda = 0;
  _img_size = cvSize(0, 0);
  _channels = 0;

  _size_pyr.clear();

  _tmp_img.reset();
  _img_pyr.clear();
  _tmp_img_pyr.clear();
  _mask_pyr.clear();
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_fill_img_pyr(CvArr const * img)
{
  {
    int const lev = 0; // HIGHEST level
    cvConvert(img, _tmp_img); // makes a copy, CONVERTS to float

    #pragma omp parallel for
    for(int y = 0; y < _img_size.height; ++y) // SPLITS CHANNELS
    {
      float const * const img_row = static_cast<float *>(_tmp_img.row(y));

      float * I_row [_channels]; // image channels
      for(int k = 0; k < _channels; ++k)
      {
        I_row[k] = static_cast<float *>(_img_pyr[lev][k].row(y));
      }

      for(int x = 0; x < _img_size.width; ++x)
      {
        for(int k = 0; k < _channels; ++k)
        {
          I_row[k][x] = img_row[_channels * x + k];
        }
      }
    }
  }

  if(_lambda == 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      for(int k = 0; k < _channels; ++k)
      {
        cvPyrDown(_img_pyr[lev-1][k], _img_pyr[lev][k]); // HALF size
      }
    }
  }
  else // if(_lambda != 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      for(int k = 0; k < _channels; ++k)
      {
        cvResize(
          _img_pyr[lev-1][k], _img_pyr[lev][k],
          //CV_INTER_NN       // Nearest-neigbor interpolation
          CV_INTER_LINEAR   // Bilinear interpolation
          //CV_INTER_AREA     // Resampling using pixel area relation
          //CV_INTER_CUBIC    // Bicubic interpolation
        );
      }
    }
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_fill_mask_pyr(CvArr const * mask)
{
  {
    int const lev = 0; // HIGHEST level
    cvCopy(mask, _mask_pyr[lev]);
  }

  if(_lambda == 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      cvPyrDown(_mask_pyr[lev-1], _mask_pyr[lev]); // HALF size
    }
  }
  else // if(_lambda != 0.5)
  {
    for(int lev = 1; lev < _lev_num; ++lev) // LOWER levels
    {
      cvResize(
        _mask_pyr[lev-1], _mask_pyr[lev],
        //CV_INTER_NN       // Nearest-neigbor interpolation
        CV_INTER_LINEAR   // Bilinear interpolation
        //CV_INTER_AREA     // Resampling using pixel area relation
        //CV_INTER_CUBIC    // Bicubic interpolation
      );
    }
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_init_data(
  CvArr const * img, CvArr const * mask,
  int lev_num, double lambda
)
{
  assert(img != 0 && mask != 0);
  assert(lev_num > 0); // AT LEAST ONE LEVEL needed
  assert(lambda > 0 && lambda < 1);

  CvSize const img_size = cvGetSize(img);

  assert(img_size.width >= MIN_FSIZE); // avoid too small images
  assert(img_size.height >= MIN_FSIZE);

  assert(cvGetSize(mask) == img_size);

  int const channels = CV_MAT_CN(cvGetElemType(img));
  assert(channels > 0); // AT LEAST ONE CHANNEL needed

  assert(cvGetElemType(mask) == CV_8UC1);
  //assert(CV_MAT_CN(cvGetElemType(mask)) == 1);

  if(
    _lev_num != lev_num || _lambda != lambda ||
    _img_size != img_size || _channels != channels
  )
  {
    _drop_data(); // DROPS OLD DATA, if there was any

    // create pyramids

    _img_pyr.resize(lev_num); // input images
    _tmp_img_pyr.resize(lev_num);

    for(int lev = 0; lev < lev_num; ++lev)
    {
      _img_pyr[lev].resize(channels);
      _tmp_img_pyr[lev].resize(channels);
    }

    _mask_pyr.resize(lev_num);

    // calculate image size on pyramid levels

    {
      _size_pyr.resize(lev_num);

      CvSize sz = img_size; // HIGHEST LEVEL
      for(int lev = 0; lev < lev_num; ++lev)
      {
        assert(sz.width >= MIN_FSIZE && sz.height >= MIN_FSIZE);
        _size_pyr[lev] = sz; // image size on current level

        if(lambda == 0.5)
        {
          sz.width /= 2, sz.height /= 2; // HALF size on next level
        }
        else // if(lambda != 0.5)
        {
          sz.width = static_cast<int>(floor(sz.width * lambda));
          sz.height = static_cast<int>(floor(sz.height * lambda));
        }
      }
    }

    // create images

    _tmp_img.reset(img_size, IPL_DEPTH_32F, channels);

    for(int lev = 0; lev < lev_num; ++lev)
    {
      CvSize const sz = _size_pyr[lev]; // current size

      for(int k = 0; k < channels; ++k)
      {
        _img_pyr[lev][k].reset(sz, IPL_DEPTH_32F, 1);
        _tmp_img_pyr[lev][k].reset(sz, IPL_DEPTH_32F, 1);
      }

      _mask_pyr[lev].reset(sz, IPL_DEPTH_8U, 1);
    }

    // store parameters of pyramids

    _lev_num = lev_num;
    _lambda = lambda;
    _img_size = img_size;
    _channels = channels;
  }

  // init image and mask pyramid

  _fill_img_pyr(img);
  _fill_mask_pyr(mask);
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_correct_boundary(
  cv::img_ptr & img, // REPLICATES VALUES on bottom-right boundary
  bool is_x_corr, bool is_y_corr
)
{
  CvSize const sz = cvGetSize(img);

  if(is_x_corr) // RIGHT column
  {
    for(int y = 0; y < sz.height; ++y)
    {
      float * const row =
        reinterpret_cast<float *>(img.row(y));

      row[sz.width-1] = row[sz.width-2];
    }
  }

  if(is_y_corr) // BOTTOM row
  {
    float * const last_row =
      reinterpret_cast<float *>(img.row(sz.height-1));
    float const * const curr_row =
      reinterpret_cast<float const *>(img.row(sz.height-2));

    memcpy(last_row, curr_row, sz.width * sizeof(float));
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_switch_to_level(int lev)
{
  assert(lev < _lev_num - 1); // FROM LOWER TO HIGHER

  if(_lambda == 0.5)
  {
    CvSize const sz = _size_pyr[lev];
    CvSize const sz_lower = _size_pyr[lev+1];

    CvSize const sz_upper =
      cvSize(2 * sz_lower.width, 2 * sz_lower.height);

    bool const is_x_corr = (sz.width != sz_upper.width);
    bool const is_y_corr = (sz.height != sz_upper.height);

    bool const is_bound_corr_needed = is_x_corr || is_y_corr;

    CvRect const rect_lower_up =
      cvRect(0, 0, sz_upper.width, sz_upper.height);

    for(int k = 0; k < _channels; ++k)
    {
      if(is_bound_corr_needed) // boundary correction?
      {
        cvSetImageROI(_tmp_img_pyr[lev][k], rect_lower_up);
      }

      cvPyrUp(_img_pyr[lev+1][k], _tmp_img_pyr[lev][k]); // size DOUBLED

      if(is_bound_corr_needed) // boundary correction?
      {
        cvResetImageROI(_tmp_img_pyr[lev][k]); // resets region of interest
        _correct_boundary(_tmp_img_pyr[lev][k], is_x_corr, is_y_corr);
      }
    }
  }
  else // if(_lambda != 0.5)
  {
    for(int k = 0; k < _channels; ++k)
    {
      cvResize(
        _img_pyr[lev+1][k], _tmp_img_pyr[lev][k],
        //CV_INTER_NN       // Nearest-neigbor interpolation
        CV_INTER_LINEAR   // Bilinear interpolation
        //CV_INTER_AREA     // Resampling using pixel area relation
        //CV_INTER_CUBIC    // Bicubic interpolation
      );
    }
  }

  for(int k = 0; k < _channels; ++k)
  {
    cvCopy(_tmp_img_pyr[lev][k], _img_pyr[lev][k], _mask_pyr[lev]);
  }
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_run_on_level(
  int lev, int it_num, double lambda, double alpha
)
{
  assert(lev < _lev_num);
  CvSize const sz = _size_pyr[lev];

  for(int step = 0; step < it_num; ++step)
  {
    #pragma omp parallel for
    for(int y = 0; y < sz.height; ++y)
    {
      float * img_row [3][_channels];

      for(int dy = -1; dy <= 1; ++dy)
      {
        int y_ = y + dy; // neighboring rows
        if(y_ < 0) y_ = 0; // REPLICATIVE BOUNDARY
        else if(y_ > sz.height - 1) y_ = sz.height - 1;

        for(int k = 0; k < _channels; ++k)
        {
          img_row[1+dy][k] =
            static_cast<float *>(_img_pyr[lev][k].row(y_));
        }
      }

      uchar const * const mask_row =
        static_cast<uchar *>(_mask_pyr[lev].row(y));

      for(int x = 0; x < sz.width; ++x)
      {
        if(mask_row[x] != 0)
        {
          for(int k = 0; k < _channels; ++k)
          {
            float * img_pix[3][3];

            for(int dx = -1; dx <= 1; ++dx)
            {
              int x_ = x + dx; // neighboring columns
              if(x_ < 0) x_ = 0; // REPLICATIVE BOUNDARY
              else if(x_ > sz.width - 1) x_ = sz.width - 1;

              for(int dy = -1; dy <= 1; ++dy)
              {
                img_pix[1+dx][1+dy] = img_row[1+dy][k] + x_;
              }
            }

            float const img_bar =
                img_pix[1+0][1-1][0]
              + img_pix[1-1][1+0][0] + img_pix[1+1][1+0][0]
              + img_pix[1+0][1+1][0]
            ;

            float * curr_pix = img_pix[1+0][1+0];

            curr_pix[0] =
              (curr_pix[0] + alpha * img_bar)
              / (1 + 4 * alpha);
          }
        }
      }
    }
  }

#if TEST_LEVEL >= MIN_TEST_LEVEL + 1
  cvNamedWindow("run");
  cvShowImage("run", _img_pyr[lev][0]);
  cvShowImage("mask", _mask_pyr[lev]);
  cvWaitKey(-1);
#endif
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::_init_output(CvArr * out)
{
  assert(out != 0 && cvGetSize(out) == _img_size);
  int const lev = 0; // HIGHEST level

  #pragma omp parallel for
  for(int y = 0; y < _img_size.height; ++y) // JOINS CHANNELS
  {
    float const * I_row [_channels]; // image channels
    for(int k = 0; k < _channels; ++k)
    {
      I_row[k] = static_cast<float const *>(_img_pyr[lev][k].row(y));
    }

    float * const out_row = static_cast<float *>(_tmp_img.row(y));

    for(int x = 0; x < _img_size.width; ++x)
    {
      float * const out_pix = out_row + _channels * x;

      for(int k = 0; k < _channels; ++k)
      {
        out_pix[k] = I_row[k][x];
      }
    }
  }

  cvConvert(_tmp_img, out); // CONVERTS from float to output type
}

//////////////////////////////////////////////////////////////////////////

void _mm_inp::inpaint(
  CvArr const * input, CvArr const * mask, CvArr * output,
  int it_num, int lev_num, double lambda, double alpha
)
{
  _init_data(input, mask, lev_num, lambda);

  int lev = _lev_num - 1; // staring on the LOWEST level

#if TEST_LEVEL >= MIN_TEST_LEVEL
  {
    CvSize const & size_k = _size_pyr[lev];
    std::cerr
      << "Level: "
      << size_k.width << " x " << size_k.height
      << std::endl;
  }
#endif

  while(lev > 0) // from lower to higher
  {
    double const alpha_ = alpha * pow(lambda, 2 * lev);
    _run_on_level(lev, it_num, lambda, alpha_); // CURRENT level

    --lev; // switching TO HIGHER LEVEL

  #if TEST_LEVEL >= MIN_TEST_LEVEL
    {
      CvSize const & size_k = _size_pyr[lev];
      std::cerr
        << "Level: "
        << size_k.width << " x " << size_k.height
        << std::endl;
    }
  #endif

    _switch_to_level(lev); // moves results to HIGHER level
  }

  _run_on_level(0, it_num, lambda, alpha); // HIGHEST level
  _init_output(output);
}

//////////////////////////////////////////////////////////////////////////
// EXPORT
//////////////////////////////////////////////////////////////////////////

mm_inpaint_func::mm_inpaint_func()
:
  _intern (0)
{
  // NO MEMORY ALLOCATED
}

//////////////////////////////////////////////////////////////////////////

mm_inpaint_func::~mm_inpaint_func()
{
  reset(); // DEALLOCATES MEMORY, if there was any allocated
}

//////////////////////////////////////////////////////////////////////////

mm_inpaint_func::mm_inpaint_func(
  mm_inpaint_func const & func
)
:
  _intern (0)
{
  // there is NO REASON TO COPY anything
  // NO MEMORY ALLOCATED
}

//////////////////////////////////////////////////////////////////////////

mm_inpaint_func &
  mm_inpaint_func::operator=(mm_inpaint_func const & func)
{
  // there is NO REASON TO COPY anything
  return *this;
}

//////////////////////////////////////////////////////////////////////////

void mm_inpaint_func::reset()
{
  if(_intern == 0) return; // IGNORED
  delete static_cast<_mm_inp *>(_intern);
  _intern = 0;
}

//////////////////////////////////////////////////////////////////////////

void mm_inpaint_func::operator()(
  CvArr const * input, CvArr const * mask, CvArr * output,
  int it_num, int lev_num, double lambda, double alpha
)
{
  if(_intern == 0)
  {
    _intern = new _mm_inp ();
    assert(_intern != 0);
  }

  _mm_inp * intern = static_cast<_mm_inp *>(_intern);

  intern->inpaint(
    input, mask, output,
    it_num, lev_num, lambda, alpha
  );
}

//////////////////////////////////////////////////////////////////////////

// End of File
