// $Id$

//////////////////////////////////////////////////////////////////////////

#include "inpaint.h"
#include "util.h"

#include <cv.h>
#include <highgui.h>
#include <imgptr.h>

//////////////////////////////////////////////////////////////////////////

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <iomanip>
#include <stdexcept>
#include <string>
#include <sstream>
#include <limits>

//////////////////////////////////////////////////////////////////////////

#ifndef TEST_LEVEL
#define TEST_LEVEL 1
#endif

#ifndef MIN_FSIZE
#define MIN_FSIZE 32
#endif

//////////////////////////////////////////////////////////////////////////

class prog_args
{
public:
  std::string in_name; // name of input file
  std::string mask_name; // mask of 'in_name'
  std::string out_name; // name of output file or directory

public:
  int it_num; // number of iteration steps
  int lev_num; // number of resoltion levels
  double lambda; // scaling factor
  double alpha; // diffusion parameter

public:
  prog_args(int argc, char * argv[]); // reads arguments
};

//////////////////////////////////////////////////////////////////////////

prog_args::prog_args(int argc, char * argv[])
{
  int const arg_num = argc - 1;
  int const arg_num_min = 7;
  int const arg_num_max = arg_num_min + 0;

  if(arg_num < arg_num_min || arg_num > arg_num_max)
  {
    std::ostringstream error;
    error
      << "arguments:"
      << " <in_name> <mask_name> <out_name>"
      << " <it_num> <lev_num> <lambda> <alpha>"
    ;
    throw std::runtime_error(error.str());
  }

  // MANDATORY ARGUMENTS

  in_name = argv[1]; // name of input file
  mask_name = argv[2]; // mask of input file
  out_name = argv[3]; // name of output file or directory

  it_num = atoi(argv[4]); // number of iteration steps

  if(it_num < 0)
  {
    std::ostringstream error;
    error << "invalid number of iterations: " << it_num;
    throw std::runtime_error(error.str());
  }

  lev_num = atoi(argv[5]); // number of iteration steps

  if(lev_num < 0)
  {
    std::ostringstream error;
    error << "invalid number of resoltion levels: " << lev_num;
    throw std::runtime_error(error.str());
  }

  lambda = atof(argv[6]); // scaling factor

  if(lambda <= 0 || lambda >= 1)
  {
    std::ostringstream error;
    error
      << "invalid scaling factor (lambda): "
      << alpha << " (should be > 0 and < 1)";
    throw std::runtime_error(error.str());
  }

  alpha = atof(argv[7]); // diffusion parameter

  if(alpha < 0)
  {
    std::ostringstream error;
    error
      << "invalid diffusion parameter (alpha): "
      << alpha << " (should be > 0)";
    throw std::runtime_error(error.str());
  }

  // OPTIONAL ARGUMENTS

  // NONE
}

//////////////////////////////////////////////////////////////////////////

class prog
{
private:
  prog_args const _args; // program arguments
  mm_inpaint_func inpaint; // multiscale-multichannel inpainting
  cv::img_ptr _in_img, _mask_img, _out_img; // input/mask/output image

public:
  prog(int argc, char * argv[])
  :
    _args (argc, argv) // reads arguments
  {
    _init(); // initializes the program
    _run(); // executes the program
  }

private:
  prog(prog const &); // NOT IMPLEMENTED
  prog& operator=(prog const &);

private:
  void _init(); // initializes the program
  void _run(); // executes the program
};

//////////////////////////////////////////////////////////////////////////

void prog::_init()
{
  // OPEN INPUT

  IplImage * in_img = cvLoadImage(_args.in_name.c_str());

  if(in_img == 0 || in_img->imageData == 0)
  {
    std::ostringstream error;
    error << "couldn't load file: " << _args.in_name;
    throw std::runtime_error(error.str());
  }
  if(in_img->nChannels != 3) // BGR
  {
    std::ostringstream error;
    error
      << "unsupported number of input channels: "
      << in_img->nChannels;
    throw std::runtime_error(error.str());
  }
  if(in_img->depth != (int) IPL_DEPTH_8U) // [0,255]
  {
    std::ostringstream error;
    error
      << "unsupported type of input data: "
      << ipl_depth_name(in_img->depth);
    throw std::runtime_error(error.str());
  }
  if(in_img->width < MIN_FSIZE || in_img->height < MIN_FSIZE)
  {
    std::ostringstream error;
    error
      << "too small input image: "
      << in_img->width << "x" << in_img->height
      << " (must be >= " << MIN_FSIZE << ")";
    throw std::runtime_error(error.str());
  }

  CvSize const in_sz = cvSize(in_img->width, in_img->height);

#if TEST_LEVEL >= 1
  std::cerr
    << "Input: " << _args.in_name << " "
    << in_sz.width << "x" << in_sz.height << " "
    << std::endl;
#endif

  {
    int w_k = in_sz.width, h_k = in_sz.height;
    for(int k = 1; k < _args.lev_num; ++k)
    {
      w_k = static_cast<int>(floor(w_k * _args.lambda));
      h_k = static_cast<int>(floor(h_k * _args.lambda));
      if(w_k < MIN_FSIZE || h_k < MIN_FSIZE)
      {
        std::ostringstream error;
        error
          << "too many levels: " << _args.lev_num
          << " (should be <= " << k << ")";
        throw std::runtime_error(error.str());
      }
    }
  }

  _in_img.reset(in_sz, CV_32FC3); // stores input

  cvConvert(in_img, _in_img); // CONVERTED to floating point
  cvScale(_in_img, _in_img, 1./255.); // SCALED to [0,1]

  cvReleaseImage(&in_img); // original released

  // READ MASK

  _mask_img = cvLoadImage(
    _args.mask_name.c_str(), CV_LOAD_IMAGE_GRAYSCALE
  );

  if(_mask_img == 0 || _mask_img->imageData == 0)
  {
    std::ostringstream error;
    error << "couldn't load mask: " << _args.mask_name;
    throw std::runtime_error(error.str());
  }

  assert(_mask_img->nChannels == 1);

  if(_mask_img->depth != (int) IPL_DEPTH_8U) // [0,255]
  {
    std::ostringstream error;
    error
      << "unsupported type of mask data: "
      << ipl_depth_name(_mask_img->depth);
    throw std::runtime_error(error.str());
  }

  if(cvGetSize(_mask_img) != in_sz)
  {
    std::ostringstream error;
    error
      << "the size of mask and image should match: "
      << _mask_img->width << "x" << _mask_img->height
      << " (must be "
      << _in_img->width << "x" << _in_img->height << ")";
    throw std::runtime_error(error.str());
  }

  cvNot(_mask_img, _mask_img);

  // OPEN OUTPUT

  CvSize const out_sz = in_sz; // output size == input size

#if TEST_LEVEL >= 1
  std::cerr
    << "Output: " << _args.out_name << " "
    << out_sz.width << "x" << out_sz.height << " "
    << std::endl;
#endif

  _out_img.reset(out_sz, CV_32FC3); // output image
}

//////////////////////////////////////////////////////////////////////////

void prog::_run()
{
#if TEST_LEVEL >= 2
  cvNamedWindow("input");
  cvShowImage("input", _in_img);

  cvNamedWindow("mask");
  cvShowImage("mask", _mask_img);
#endif

#if TEST_LEVEL >= 2
  std::cerr << "Press any key to start!" << std::endl;
  cvWaitKey(-1); // WAITS FOR ANY KEY
  std::cerr << "Running ..." << std::endl;
#endif

  inpaint(
    _in_img, _mask_img, _out_img, // multiscale multichannel inpainting
    _args.it_num, _args.lev_num, _args.lambda, _args.alpha
  );

#if 0 // HACK
  cv::img_ptr texture (_in_img);
  cvSmooth(texture, texture, CV_GAUSSIAN, 5);
  cvSub(_in_img, texture, texture);
  cvAdd(_out_img, texture, _out_img, _mask_img);
#endif

  cvSaveImage(_args.out_name.c_str(), _out_img);

#if TEST_LEVEL >= 2
  cvNamedWindow("result");
  cvShowImage("result", _out_img);
  std::cerr << "Press any key ..." << std::endl;
  cvWaitKey(-1);
#endif
}

//////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv[])
{
  std::string const prog_name = argv[0];
  try
  {
    prog(argc, argv);
    return EXIT_SUCCESS;
  }
  catch(std::exception const & e)
  {
    std::cerr
      << prog_name << ": " << e.what()
      << std::endl;
  }
  return EXIT_FAILURE;
}

//////////////////////////////////////////////////////////////////////////

// End of File
