#ifndef SATUCAST_H
#define SATUCAST_H

// Copyleft (x) Sandor Fazekas, 2008. All rights reversed.
// $Id: satucast.h,v 1.2 2008/07/13 20:53:01 fazekas Exp $

//////////////////////////////////////////////////////////////////////////

#include <limits>

//////////////////////////////////////////////////////////////////////////

template<int code>
struct _saturate_caster
{
  // SEE SPECILIZATIONS
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _saturate_caster<0> // SIMPLE CAST
{
  template<typename D, typename S>
  inline static void convert(S const & src, D & dst)
  {
    dst = static_cast<D>(src);
  }
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _saturate_caster<1> // MIN/MAX LIMIT
{
  template<typename D, typename S>
  inline static void convert(S const & src, D & dst)
  {
    D const d_min = std::numeric_limits<D>::min();
    D const d_max = std::numeric_limits<D>::max();
    S const d_min_s = static_cast<S>(d_min);
    S const d_max_s = static_cast<S>(d_max);
    if(src <= d_min_s) dst = d_min;
    else if(src >= d_max_s) dst = d_max;
    else dst = static_cast<D>(src);
  }
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _saturate_caster<2> // MAX LIMIT
{
  template<typename D, typename S>
  inline static void convert(S const & src, D & dst)
  {
    D const d_max = std::numeric_limits<D>::max();
    S const d_max_s = static_cast<S>(d_max);
    if(src >= d_max_s) dst = d_max;
    else dst = static_cast<D>(src);
  }
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _saturate_caster<3> // 0 MIN LIMIT
{
  template<typename D, typename S>
  inline static void convert(S const & src, D & dst)
  {
    if(src <= 0) dst = 0;
    else dst = static_cast<D>(src);
  }
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _saturate_caster<4> // MIN/MAX LIMIT ROUND
{
  template<typename D, typename S>
  inline static void convert(S const & src, D & dst)
  {
    D const d_min = std::numeric_limits<D>::min();
    D const d_max = std::numeric_limits<D>::max();
    S const d_min_s = static_cast<S>(d_min);
    S const d_max_s = static_cast<S>(d_max);
    S const src_ = src + 0.5; // src is FLOATING POINT
    if(src_ <= d_min_s) dst = d_min;
    else if(src_ >= d_max_s) dst = d_max;
    else dst = static_cast<D>(src_);
  }
};

//////////////////////////////////////////////////////////////////////////

template<typename D, typename S, int round = 0>
struct _saturate_caster_rule
{
  enum {
    code =
      std::numeric_limits<D>::is_integer ?
      (
        std::numeric_limits<S>::is_integer ?
        (
          std::numeric_limits<D>::is_signed ?
          (
            std::numeric_limits<S>::is_signed ?
            (
              sizeof(D) >= sizeof(S) ?
              (
                0 // SIMPLE CAST
              )
              : // sizeof(D) < sizeof(S) ?
              (
                1 // MIN/MAX LIMIT
              )
            )
            : // !std::numeric_limits<S>::is_signed ?
            (
              sizeof(D) > sizeof(S) ?
              (
                0 // SIMPLE CAST
              )
              : // sizeof(D) <= sizeof(S) ?
              (
                2 // MAX LIMIT
              )
            )
          )
          : // !std::numeric_limits<D>::is_signed ?
          (
            std::numeric_limits<S>::is_signed ?
            (
              sizeof(D) >= sizeof(S) ?
              (
                3 // 0 MIN LIMIT
              )
              : // sizeof(D) < sizeof(S) ?
              (
                1 // MIN/MAX LIMIT
              )
            )
            : // !std::numeric_limits<S>::is_signed ?
            (
              sizeof(D) >= sizeof(S) ?
              (
                0 // SIMPLE CAST
              )
              : // sizeof(D) < sizeof(S) ?
              (
                2 // MAX LIMIT
              )
            )
          )
        )
        : // !std::numeric_limits<S>::is_integer ?
        (
          (round == 0) ? 1 : 4 // MIN/MAX LIMIT (ROUND)
        )
      )
      : // !std::numeric_limits<D>::is_integer ?
      (
        0 // SIMPLE CAST
      )
  };
};

//////////////////////////////////////////////////////////////////////////

template<typename D, typename S>
inline D saturate_cast(S const src)
{
  D dst;
  _saturate_caster<
      _saturate_caster_rule<D, S>::code
    >::convert(src, dst);
  return dst;
}

//////////////////////////////////////////////////////////////////////////

template<typename D, typename S>
inline D saturate_round_cast(S const src)
{
  D dst;
  _saturate_caster<
      _saturate_caster_rule<D, S, 1>::code
    >::convert(src, dst);
  return dst;
}

//////////////////////////////////////////////////////////////////////////

#endif // SATUCAST_H
