// Copyleft (x) Sandor Fazekas, 2008. All rights reversed.
// $Id: unwarp.cpp,v 1.2 2008/07/13 20:37:01 fazekas Exp $

//////////////////////////////////////////////////////////////////////////

#include "unwarp.h"

#include <cv.h>
#include <satucast.h>

#include <cassert>
#include <limits>

//////////////////////////////////////////////////////////////////////////
namespace _unwarp_ {
//////////////////////////////////////////////////////////////////////////

typedef unsigned char   u8;   // unsigned 8-bit integers
typedef signed char     s8;   // signed 8-bit integers
typedef unsigned short  u16;  // unsigned 16-bit integers
typedef signed short    s16;  // signed 16-bit integers
//typedef unsigned int    u32;  // unsigned 32-bit integers: UNSUPPORTED
typedef signed int      s32;  // signed 32-bit integers
typedef float           f32;  // single precision floating-point numbers
typedef double          f64;  // double precision floating-point numbers

//////////////////////////////////////////////////////////////////////////

template<int code>
struct ipl_type
{
  typedef void type; // INVALID!
};

template<typename T>
struct ipl_depth
{
  static int const code = 0; // INVALID!
};

#define _DECLARE_IPL_TYPE_DEPTH(T, CODE) \
template<> \
struct ipl_type<CODE> \
{ \
  typedef T type; \
}; \
\
template<> \
struct ipl_depth<T> \
{ \
  static int const code = CODE; \
};

_DECLARE_IPL_TYPE_DEPTH(u8, IPL_DEPTH_8U)
_DECLARE_IPL_TYPE_DEPTH(s8, IPL_DEPTH_8S)
_DECLARE_IPL_TYPE_DEPTH(u16, IPL_DEPTH_16U)
_DECLARE_IPL_TYPE_DEPTH(s16, IPL_DEPTH_16S)
//_DECLARE_IPL_TYPE_DEPTH(u32, IPL_DEPTH_32U)
_DECLARE_IPL_TYPE_DEPTH(s32, IPL_DEPTH_32S)
_DECLARE_IPL_TYPE_DEPTH(f32, IPL_DEPTH_32F)
_DECLARE_IPL_TYPE_DEPTH(f64, IPL_DEPTH_64F)

//////////////////////////////////////////////////////////////////////////

template<typename T>
struct ipl_depth_signed
{
  static int const code = ipl_depth<T>::code;
};

#define _DECLARE_IPL_DEPTH_SIGNED(T, CODE) \
template<> \
struct ipl_depth_signed<T> \
{ \
  static int const code = CODE; \
};

_DECLARE_IPL_DEPTH_SIGNED(u8, ipl_depth<s8>::code)
_DECLARE_IPL_DEPTH_SIGNED(u16, ipl_depth<s16>::code)
//_DECLARE_IPL_DEPTH_SIGNED(u32, ipl_depth<s32>::code)

template<typename T>
struct ipl_signed
{
  typedef typename ipl_type<
      ipl_depth_signed<T>::code
    >::type type;
};

//////////////////////////////////////////////////////////////////////////

template<typename T1, typename T2>
struct ipl_depth_fusion
{
  static int const code =
  (
    !std::numeric_limits<T1>::is_integer &&
    std::numeric_limits<T2>::is_integer
  )
  ? ipl_depth<T1>::code // FLOATING POINT
  : (
      (
        std::numeric_limits<T1>::is_integer &&
        !std::numeric_limits<T2>::is_integer
      )
      ? ipl_depth<T2>::code // FLOATING POINT
      : (
          (
            !std::numeric_limits<T1>::is_integer &&
            !std::numeric_limits<T2>::is_integer
          )
          ? (
              (sizeof(T1) > sizeof(T2))
              ? ipl_depth<T1>::code // FLOATING POINT
              : ipl_depth<T2>::code // FLOATING POINT
            )
          : (
              // std::numeric_limits<T1>::is_integer &&
              // std::numeric_limits<T2>::is_integer
              (sizeof(T1) > sizeof(T2))
              ? (
                  (std::numeric_limits<T1>::is_signed)
                  ? ipl_depth<T1>::code
                  : (
                      (std::numeric_limits<T2>::is_signed)
                      ? ipl_depth_signed<T1>::code
                      : ipl_depth<T1>::code
                    )
                )
              : (
                  (sizeof(T1) < sizeof(T2))
                  ? (
                      (std::numeric_limits<T2>::is_signed)
                      ? ipl_depth<T2>::code
                      : (
                          (std::numeric_limits<T1>::is_signed)
                          ? ipl_depth_signed<T2>::code
                          : ipl_depth<T2>::code
                        )
                    )
                  : (
                      // sizeof(T1) == sizeof(T2)
                      (std::numeric_limits<T1>::is_signed)
                      ? ipl_depth<T1>::code
                      : ipl_depth<T2>::code
                    )
                )
            )
        )
    );
};

//////////////////////////////////////////////////////////////////////////

template<typename T1, typename T2>
struct ipl_fusion
{
  typedef typename ipl_type<
      ipl_depth_fusion<T1, T2>::code
    >::type type;
};

//////////////////////////////////////////////////////////////////////////
} // namespace _unwarp_
//////////////////////////////////////////////////////////////////////////

using namespace _unwarp_;

//////////////////////////////////////////////////////////////////////////

template<int mode>
struct _calc_unwarp // INTEGER or FLOATING POINT displacements
{
  // SEE SPECIALIZATIONS
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _calc_unwarp<0> // INTEGER displacements
{
  template<typename T, typename F>
  static void run(
    CvMat const & src, CvMat & dst, CvMat const & dx, CvMat const & dy,
    int const row_num, int const col_num, int const ch_num
  )
  {
    assert(std::numeric_limits<F>::is_integer);

    #pragma omp parallel for
    for(int y = 0; y < row_num; ++y)
    {
      F const * const dx_row =
        reinterpret_cast<F const *>(dx.data.ptr + y * dx.step);
      F const * const dy_row =
        reinterpret_cast<F const *>(dy.data.ptr + y * dy.step);

      T * const dst_row =
        reinterpret_cast<T *>(dst.data.ptr + y * dst.step);

      for(int x = 0; x < col_num; ++x)
      {
        int const u = static_cast<int>(dx_row[x]);
        int const v = static_cast<int>(dy_row[x]);

        int const x_ = x + u; // SOURCE pixel
        int const y_ = y + v;

        T const * const src_row =
          reinterpret_cast<T const *>(
            (y_ > 0)
            ? (y_ < row_num)
              ? (src.data.ptr + y_ * src.step)
              : (src.data.ptr + (row_num - 1) * src.step)
            : src.data.ptr
          );

        T const * const src_pix = src_row +
          (
            (x_ > 0)
            ? (x_ < col_num)
              ? (ch_num * x_)
              : (ch_num * (col_num - 1))
            : 0
          );

        T * const dst_pix = dst_row + ch_num * x;

        for(int k = 0; k < ch_num; ++k) // FOR EACH CHANNEL
        {
          dst_pix[k] = src_pix[k];
        }
      }
    }
  }
};

//////////////////////////////////////////////////////////////////////////

template<>
struct _calc_unwarp<1> // FLOATING POINT displacements
{
  template<typename T, typename F>
  static void run(
    CvMat const & src, CvMat & dst, CvMat const & dx, CvMat const & dy,
    int const row_num, int const col_num, int const ch_num
  )
  {
    assert(!std::numeric_limits<F>::is_integer);
    typedef typename ipl_fusion<T, F>::type A;

    #pragma omp parallel for
    for(int y = 0; y < row_num; ++y)
    {
      F const * const dx_row =
        reinterpret_cast<F const *>(dx.data.ptr + y * dx.step);
      F const * const dy_row =
        reinterpret_cast<F const *>(dy.data.ptr + y * dy.step);

      T * const dst_row =
        reinterpret_cast<T *>(dst.data.ptr + y * dst.step);

      for(int x = 0; x < col_num; ++x)
      {
        A const u = static_cast<A>(dx_row[x]);
        A const v = static_cast<A>(dy_row[x]);

        A const u_int_ = static_cast<A>(floor(u));
        A const v_int_ = static_cast<A>(floor(v));

        int const u_int = static_cast<int>(u_int_);
        int const v_int = static_cast<int>(v_int_);

        A const u_frac = u - u_int_; // FRACTIONAL part
        A const v_frac = v - v_int_; // 0 <= {u,v}_frac < 1.0

        A const weight11 = u_frac * v_frac;
        A const weight10 = u_frac - weight11;
        A const weight01 = v_frac - weight11;
        A const weight00 = 1 - u_frac - v_frac + weight11;

        int const x_ = x + u_int; // SOURCE pixel
        int const y_ = y + v_int;

        uchar const * const src_row0_ =
          (y_ > 0)
          ? (y_ < row_num)
            ? src.data.ptr + y_ * src.step
            : src.data.ptr + (row_num - 1) * src.step
          : src.data.ptr;

        uchar const * const src_row1_ =
          (y_ + 1 > 0)
          ? (y_ + 1 < row_num)
            ? src_row0_ + src.step
            : src_row0_
          : src.data.ptr;

        T const * const src_row0 =
          reinterpret_cast<T const *>(src_row0_);
        T const * const src_row1 =
          reinterpret_cast<T const *>(src_row1_);

        int const src_ind0 =
          (x_ > 0)
          ? (x_ < col_num)
            ? ch_num * x_
            : ch_num * (col_num - 1)
          : 0;

        int const src_ind1 =
          (x_ + 1 > 0)
          ? (x_ + 1 < col_num)
            ? src_ind0 + ch_num
            : src_ind0
          : 0;

        T const * const src_pix00 = src_row0 + src_ind0;
        T const * const src_pix01 = src_row1 + src_ind0;
        T const * const src_pix10 = src_row0 + src_ind1;
        T const * const src_pix11 = src_row1 + src_ind1;

        T * const dst_pix = dst_row + ch_num * x;

        for(int k = 0; k < ch_num; ++k)
        {
          A const src00 = static_cast<A>(src_pix00[k]);
          A const src01 = static_cast<A>(src_pix01[k]);
          A const src10 = static_cast<A>(src_pix10[k]);
          A const src11 = static_cast<A>(src_pix11[k]);

          A const src_unwarped =
            src00 * weight00 + src01 * weight01 +
            src10 * weight10 + src11 * weight11;

          dst_pix[k] = saturate_round_cast<T>(src_unwarped);
        }
      }
    }
  }
};

//////////////////////////////////////////////////////////////////////////

void cvExUnwarp(
  CvArr const * src, CvArr * dst,
  CvArr const * dx, CvArr const * dy
)
{
  assert(src != 0 && dst != 0 && dx != 0 && dy != 0);

  CvMat src_mat, dst_mat, dx_mat, dy_mat; // IplImage -> CvMat
  int src_coi, dst_coi, dx_coi, dy_coi; // channel of interest

  CvMat const * const src_ = cvGetMat(src, &src_mat, &src_coi);
  CvMat * const dst_ = cvGetMat(dst, &dst_mat, &dst_coi);
  CvMat const * const dx_ = cvGetMat(dx, &dx_mat, &dx_coi);
  CvMat const * const dy_ = cvGetMat(dy, &dy_mat, &dy_coi);

  assert(src_ != 0 && dst_ != 0 && dx_ != 0 && dy_ != 0);
  assert(!src_coi && !dst_coi && !dx_coi && !dy_coi); // no COI!

  assert(CV_ARE_SIZES_EQ(src_, dst_));
  assert(CV_ARE_SIZES_EQ(src_, dx_));
  assert(CV_ARE_SIZES_EQ(dx_, dy_));

  assert(CV_ARE_CNS_EQ(src_, dst_));
  assert(CV_MAT_CN(dx_->type) == 1);
  assert(CV_ARE_CNS_EQ(dy_, dx_));

  int const row_num = src_->rows;
  int const col_num = src_->cols;
  int const ch_num = CV_MAT_CN(src_->type);

  assert(row_num > 0 && col_num > 0);
  assert(ch_num > 0);

  assert(CV_ARE_TYPES_EQ(src_, dst_));
  assert(CV_ARE_TYPES_EQ(dx_, dy_));

  int const src_mat_type = CV_MAT_TYPE(src_->type);
  int const src_depth = cvCvToIplDepth(src_mat_type);

  int const displ_mat_type = CV_MAT_TYPE(dx_->type);
  int const displ_depth = cvCvToIplDepth(displ_mat_type);

  #define _UNWARP_CALC_TF(T, F) \
  { \
    _calc_unwarp< \
        std::numeric_limits<F>::is_integer ? 0 : 1 \
      >::run<T, F>( \
        *src_, *dst_, *dx_, *dy_, \
        row_num, col_num, ch_num \
      ); \
  }

  #define _UNWARP_CALC_T(T) \
  { \
    switch(displ_depth) \
    { \
      case ipl_depth<u8> ::code: _UNWARP_CALC_TF(T, u8 ); break; \
      case ipl_depth<s8> ::code: _UNWARP_CALC_TF(T, s8 ); break; \
      case ipl_depth<u16>::code: _UNWARP_CALC_TF(T, u16); break; \
      case ipl_depth<s16>::code: _UNWARP_CALC_TF(T, s16); break; \
      case ipl_depth<s32>::code: _UNWARP_CALC_TF(T, s32); break; \
      case ipl_depth<f32>::code: _UNWARP_CALC_TF(T, f32); break; \
      case ipl_depth<f64>::code: _UNWARP_CALC_TF(T, f64); break; \
      default: assert(false && "unsupported displacement depth"); \
    } \
  }

  switch(src_depth)
  {
    case ipl_depth<u8> ::code: _UNWARP_CALC_T(u8);  break;
    case ipl_depth<s8> ::code: _UNWARP_CALC_T(s8);  break;
    case ipl_depth<u16>::code: _UNWARP_CALC_T(u16); break;
    case ipl_depth<s16>::code: _UNWARP_CALC_T(s16); break;
    case ipl_depth<s32>::code: _UNWARP_CALC_T(s32); break;
    case ipl_depth<f32>::code: _UNWARP_CALC_T(f32); break;
    case ipl_depth<f64>::code: _UNWARP_CALC_T(f64); break;
    default: assert(false && "unsupported source depth");
  }
}

//////////////////////////////////////////////////////////////////////////

// End of File
